package Evo::Export::Exporter;
use Evo;
use Evo::Util;
use Carp 'croak';
use Module::Load 'load';

# most of the method accept hash with key src(module), name(method), orig_src, orig_name, dst

our @CARP_NOT = ('Evo::Util');
sub new { bless {data => {}}, __PACKAGE__ }

sub data { shift->{data} }

sub install {
  my ($self, $src, $dst, @xlist) = @_;
  my @list = $self->expand_wildcards($src, @xlist);

  my %patch;
  foreach my $name_as (@list) {
    my ($name, $as) = split ':', $name_as;
    $as ||= $name;
    my $fn = $self->request_gen($src, $name, $dst);
    $patch{$as} = $fn;
  }

  Evo::Util::monkey_patch $dst, %patch;
}

sub request_gen {
  my ($self, $epkg, $name, $dpkg) = @_;
  my $slot = $self->find_slot($epkg, $name);

  # it's important to return same function for the same module
  return $slot->{cache}{$dpkg} if $slot->{cache}{$dpkg};
  return $slot->{cache}->{$dpkg} = $slot->{gen}->($dpkg);
}

# traverse to find gen via links, return Module, name, gen
sub find_slot {
  my ($self, $pkg, $name) = @_;
  croak qq{"$pkg" doesn't export "$name"} unless my $slot = $self->data->{$pkg}{$name};
}

sub init_slot {
  my ($self, $src, $name, $val) = @_;
  croak "$src already exports $name" if $self->data->{$src}{$name};
  $self->data->{$src}{$name} = $val;
}

sub add_proxy {
  my ($self, $epkg, $ename, $spkg, $sname) = @_;
  my $slot = $self->find_slot($spkg, $sname);

  $self->init_slot($epkg, $ename, $slot);
}

sub add_gen {
  my ($self, $src, $name, $gen) = @_;
  $self->init_slot($src, $name, {gen => $gen});
}

sub add_sub {
  my ($self, $src, $name_as) = @_;
  my ($name, $as) = split ':', $name_as;
  $as ||= $name;
  my $full = "${src}::$name";
  no strict 'refs';    ## no critic
  my $sub = *{$full}{CODE} or croak "Subroutine $full doesn't exists";
  $self->add_gen($src, $as, sub {$sub});
}

sub proxy {
  my ($self, $epkg, $spkg, @xlist) = @_;
  $spkg = Evo::Util::resolve_package(scalar caller, $spkg);
  load $spkg;
  my @list = $self->expand_wildcards($spkg, @xlist);

  foreach my $name_as (@list) {
    my ($sname, $ename) = split ':', $name_as;
    $ename ||= $sname;
    $self->add_proxy($epkg, $ename, $spkg, $sname);
  }

}

sub expand_wildcards {
  my ($self, $src, @list) = @_;
  croak "Empty list" unless @list;

  my $data = $self->data->{$src};
  my (%minus, %res);
  foreach my $cur (@list) {
    if ($cur eq '*') {
      croak "$src exports nothing, can't expand *" unless $data;
      $res{$_}++ for keys %$data;
    }
    elsif ($cur =~ /^-(.+)/) {
      $minus{$1}++;
    }
    else {
      $res{$cur}++;
    }
  }
  return (sort grep { !$minus{$_} } keys %res);
}


1;

__END__

=pod

=encoding UTF-8

=head1 NAME

Evo::Export::Exporter

=head1 VERSION

version 0.0173

=head1 AUTHOR

alexbyk.com

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2015 by alexbyk.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
